#-------------------------------------------------------------------------------
# Copyright (c) 2012 University of Illinois, NCSA.
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the
# University of Illinois/NCSA Open Source License
# which accompanies this distribution, and is available at
# http://opensource.ncsa.illinois.edu/license.html
#-------------------------------------------------------------------------------

##--------------------------------------------------------------------------------------------------#
##'
##' Performs an arithmetic function, FUN, over a series of traits and returns
##' the result as a derived trait.
##' Traits must be specified as either lists or single row data frames,
##' and must be either single data points or normally distributed.
##' In the event one or more input traits are normally distributed,
##' the resulting distribution is approximated by numerical simulation.
##' The output trait is effectively a copy of the first input trait with
##' modified mean, stat, and n.
##'
##' @name derive.trait
##' @title Performs an arithmetic function, FUN, over a series of traits and returns the result as a derived trait.
##' @param FUN arithmetic function
##' @param ... traits that will be supplied to FUN as input
##' @param input list of trait inputs. See examples
##' @param var.name name to use in output
##' @param sample.size number of random samples generated by rnorm for normally distributed trait input
##' @return a copy of the first input trait with mean, stat, and n reflecting the derived trait
##' @export
##' @examples
##' input <- list(x = data.frame(mean = 1, stat = 1, n = 1))
##' derive.trait(FUN = identity, input = input, var.name = 'x')
derive.trait <- function(FUN, ..., input = list(...), var.name = NA, sample.size = 10^6){
  if(any(lapply(input, nrow) > 1)){
    return(NULL)
  }
  input.samples  <- lapply(input, take.samples, sample.size=sample.size)
  output.samples <- do.call(FUN, input.samples)
  output         <- input[[1]]
  output$mean    <- mean(output.samples)
  output$stat    <- ifelse(length(output.samples) > 1, stats::sd(output.samples), NA)
  output$n       <- min(sapply(input, function(trait){trait$n}))
  output$vname   <- ifelse(is.na(var.name), output$vname, var.name)
  return(output)
}
##==================================================================================================#