#-------------------------------------------------------------------------------
# Copyright (c) 2012 University of Illinois, NCSA.
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the 
# University of Illinois/NCSA Open Source License
# which accompanies this distribution, and is available at
# http://opensource.ncsa.illinois.edu/license.html
#-------------------------------------------------------------------------------

##--------------------------------------------------------------------------------------------------#
##' Convert BioCro output to netCDF
##'
##' Converts BioCro output to netCDF.
##' Modified from on model2netcdf.SIPNET and model2netcdf.ED2 by
##' @name model2netcdf.BIOCRO
##' @title Function to convert biocro model output to standard netCDF format
##' @param result a dataframe of model output to be converted
##' @param genus character: What kind of plant was being simulated?
##'   Used to correct for some genus-specific differences in model output;
##'   Eventually that will be handled inside BioCro and this argument will be removed.
##' @param outdir Location of model output
##' @param lat Latitude of the site
##' @param lon Longitude of the site
##' @export
##' @author David LeBauer, Deepak Jaiswal, Rob Kooper
model2netcdf.BIOCRO <- function(result, genus = NULL, outdir, lat = -9999, lon = -9999) {

  if (!("hour" %in% colnames(result))) {
    result$hour <- 0
  }
  if (all(c("year", "hour", "doy") %in% colnames(result))) {
    data.table::setnames(result, c("year", "hour", "doy"), c("Year", "Hour", "DayofYear"))
  }
  
  ## longname prefix station_* used for a point
  ## http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.4/cf-conventions.html#scalar-coordinate-variables
  x <- ncdf4::ncdim_def("latitude", "degrees_north", 
                 vals = as.numeric(lat),
                 longname = "station_latitude", 
                 unlim = TRUE)
  y <- ncdf4::ncdim_def("longitude", "degrees_east",
                 vals = as.numeric(lon), 
                 longname = "station_longitude", 
                 unlim = TRUE)
  
  for (yeari in unique(result$Year)) {
    result_yeari <- result[result$Year == yeari]
    dates <- lubridate::ymd(paste0(result_yeari$Year, "-01-01")) + lubridate::days(as.numeric(result_yeari$DayofYear - 1)) + 
      lubridate::hours(result_yeari$Hour)
    days_since_origin <- dates - lubridate::ymd_hms("1700-01-01 00:00:00")
    if (!units(days_since_origin) == "days") {
      stop("check time units")
    }
    t <- ncdf4::ncdim_def("time", "days since 1700-01-01", as.numeric(days_since_origin))  # define netCDF dimensions for variables
    if (exists("genus") & (genus == "Saccharum")) {
      for (variable in c("Leaf", "Root", "Stem", "LAI", "DayofYear")) {
        v <- result_yeari[[variable]]
        result_yeari[[variable]] <- c(v[1], rep(v[-1], 24, each = TRUE))
      }
    }
   
    dims <- list(lat = x, lon = y, time = t)
    vars <- list(NPP = PEcAn.utils::to_ncvar("NPP", dims),
                 TotLivBiom = PEcAn.utils::to_ncvar("TotLivBiom", dims),
                 root_carbon_content = PEcAn.utils::to_ncvar("root_carbon_content", dims),
                 AbvGrndWood = PEcAn.utils::to_ncvar("AbvGrndWood", dims),
                 AGB = PEcAn.utils::to_ncvar("AGB", dims),
                 Evap = PEcAn.utils::to_ncvar("Evap", dims),
                 TVeg = PEcAn.utils::to_ncvar("TVeg", dims),
                 LAI = PEcAn.utils::to_ncvar("LAI", dims))
    
    biomass2c <- 0.4
    k <- udunits2::ud.convert(1, "Mg/ha", "kg/m2") * biomass2c
    
    result_yeari_std <- with(result_yeari, list(
      TotLivBiom = k * (Leaf + Root + Stem + Rhizome + Grain), 
      root_carbon_content = k * Root,
      AbvGrndWood = k * Stem,
      AGB =  k * (Leaf + Stem + Grain),
      Evap = udunits2::ud.convert(SoilEvaporation + CanopyTrans, "Mg/ha/h", "kg/m2/s"), 
      TVeg = udunits2::ud.convert(CanopyTrans, "Mg/ha/h", "kg/m2/s"), 
      LAI = LAI))
    
    total_biomass <- with(result_yeari, # this is for calculating NPP and includes litter
                          k * (Leaf + Root + Stem + Rhizome + Grain + AboveLitter + BelowLitter))
    delta_biomass <- udunits2::ud.convert(c(0, diff(total_biomass)), "kg/m2/h", "kg/m2/s")
    delta_biomass[delta_biomass < 0] <- 0
    result_yeari_std$NPP <- delta_biomass
    ncfile <- file.path(outdir, paste0(yeari, ".nc"))
    if (file.exists(ncfile)) {
      nc <- ncdf4::nc_open(ncfile, write = TRUE)
    } else {
      nc <- ncdf4::nc_create(filename = file.path(outdir, paste0(yeari, ".nc")), vars = vars)
      ncdf4::ncatt_put(nc, 0, "description", "This is an output from the BioCro Crop model generated by the model2netcdf.BIOCRO.R function in the PEcAn.BIOCRO package; see https://pecanproject.github.io/pecan-documentation/master/ for more information")
    }
    
    varfile <- file(file.path(outdir, paste(yeari, "nc", "var", sep = ".")), "w")
    
    ## Output netCDF data
    for (.vname in names(vars)) {
      ncdf4::ncvar_put(nc, varid = vars[[.vname]], vals = result_yeari_std[[.vname]])
      cat(paste(vars[[.vname]]$name, vars[[.vname]]$longname), file = varfile, 
          sep = "\n")
    }
    
    close(varfile)
    ncdf4::nc_close(nc)
  }
}  # model2netcdf.BIOCRO
