#' Write ED2IN list to file
#'
#' This writes a ED2IN file from an `ed2in` list. Default method writes a 
#' barebones file without comments. S3 method for `ed2in` objects extracts 
#' comments and their locations from the object attributes (if `barebones` is 
#' `FALSE`).
#'
#' @param ed2in Named list of ED2IN tag-value pairs. See [read_ed2in].
#' @param filename Target file name
#' @param custom_header Character vector for additional header comments. Each 
#' item gets its own line.
#' @param barebones Logical. If `TRUE`, omit comments and only write tag-value pairs.
#' @export
write_ed2in <- function(ed2in, filename, custom_header = character(), barebones = FALSE) {
  UseMethod("write_ed2in", ed2in)
}

#' @rdname write_ed2in
#' @export
write_ed2in.ed2in <- function(ed2in, filename, custom_header = character(), barebones = FALSE) {
  tags_values_vec <- tags2char(ed2in)
  if (isTRUE(barebones)) {
    write_ed2in.default(ed2in, filename, custom_header, barebones)
    return(NULL)
  }
  nvalues <- length(tags_values_vec)
  ncomments <- length(attr(ed2in, "comment_values"))
  file_body <- character(nvalues + ncomments)
  file_body[attr(ed2in, "comment_linenos")] <- attr(ed2in, "comment_values")
  file_body[attr(ed2in, "value_linenos")] <- tags_values_vec
  header <- c(
    "!=======================================",
    "!=======================================",
    "!  ED2 namelist file",
    "!  Generated by `PEcAn.ED2::write_ed2in.ed2in`",
    "!  Additional user comments below: ",
    paste0("!   ", custom_header),
    "!---------------------------------------"
  )
  output_lines <- c(header, file_body)
  writeLines(output_lines, filename)
}

#' @rdname write_ed2in
#' @export
write_ed2in.default <- function(ed2in, filename, custom_header = character(), barebones = FALSE) {
  tags_values_vec <- tags2char(ed2in)
  header <- c(
    "!=======================================",
    "!=======================================",
    "!  ED2 namelist file",
    "!  Generated by `PEcAn.ED2::write_ed2in.default`",
    "!  Additional user comments below: ",
    paste0("!   ", custom_header),
    "!---------------------------------------"
  )
  output_lines <- c(header, "$ED_NL", tags_values_vec, "$END")
  writeLines(output_lines, filename)
}

#' Format ED2IN tag-value list
#'
#' Converts an `ed2in`-like list to an ED2IN-formatted character vector.
#'
#' @inheritParams write_ed2in
tags2char <- function(ed2in) {
  char_values <- vapply(ed2in, is.character, logical(1))
  na_values <- vapply(ed2in, function(x) all(is.na(x)), logical(1))
  quoted_vals <- ed2in
  quoted_vals[char_values] <- lapply(quoted_vals[char_values], shQuote)
  quoted_vals[na_values] <- lapply(quoted_vals[na_values], function(x) "")
  values_vec <- vapply(quoted_vals, paste, character(1), collapse = ",")
  tags_values_vec <- sprintf("   NL%%%s = %s", names(values_vec), values_vec)
  tags_values_vec
}
