#-------------------------------------------------------------------------------
# Copyright (c) 2012 University of Illinois, NCSA.
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the
# University of Illinois/NCSA Open Source License
# which accompanies this distribution, and is available at
# http://opensource.ncsa.illinois.edu/license.html
#-------------------------------------------------------------------------------

##' Trait Meta-analysis
##'
##' Runs heirarchical meta-analysis of plant trait data
##'
##' `pecan.ma` runs a hierarchical Bayesian meta-analytical model.
##' This model combines prior information with data from studies on the particular species or group of interest.
##' Data that is incorporated into the meta-analysis include the mean (`Y`), sample size (`n`),
##' and precision (`obs.prec`).
##' Where a set of data includes more than one level of treatment, comes from more than one site,
##' or comes from both field and greenhouse studies, these variables are included as random (`treatment`, `site`)
##' or fixed (`greenhouse`) effects.
##' The `pecan.ma` function writes a model for each specific data set and prior using the [write.ma.model()]
##' function to modify the `ma.model.template.bug` generic model.
##'
##' @param trait.data list of `data.frame`s, one per trait for which
##'   data is available, generated by call to
##'   [PEcAn.DB::query.traits()], and post-processed by [jagify()].
##' @param prior.distns `data.frame` of prior distributions generated
##'   by call to [PEcAn.DB::query.priors()]
##' @param taupriors priors on variance parameters, can be scaled as needed with data mean
##' @param j.iter number of MCMC samples
##' @param outdir output directory
##' @param random use random effects, `FALSE` by default
##' @param overdispersed `TRUE` by default, if set to `FALSE`, data mean will be used as starting point for MCMC chains (use with caution)
##' @param logfile Path to file for sinking meta analysis output. If
##'   `NULL`, only print output to console.
##' @param verbose Logical. If `TRUE` (default), print progress messages.
##' @return four chains with 5000 total samples from posterior
##' @author David LeBauer, Michael C. Dietze, Alexey Shiklomanov
##' @export
##' @examples
##' \dontrun{
##'   # Setup
##'   con <- PEcAn.DB::db.open(...)
##'   pft <- "temperate.Early_Hardwood"
##'   pft_id <- PEcAn.DB::db.query("SELECT id FROM pfts WHERE name = $1", con,
##'   values = list(pft))[[1]]
##'   traits <- c("SLA", "Vcmax")
##'   trait_string <- paste(shQuote(traits), collapse = ",")
##'   
##'   # Load traits and priors from BETY
##'   species <- PEcAn.DB::query.pft_species(pft, con = con)
##'   trait.data <- PEcAn.DB::query.traits(species[["id"]], c("SLA", "Vcmax"), con = con)
##'   prior.distns <- PEcAn.DB::query.priors(pft_id, trait_string, con = con)
##'   
##'   # Pre-process data
##'   jagged.data <- lapply(trait.data, PEcAn.MA::jagify)
##'   taupriors <- list(tauA = 0.01,
##'   tauB = c(SLA = 1000, Vcmax = 1000))
##'   result <- pecan.ma(jagged.data, prior.distns, taupriors,
##'                      j.iter = 5000, outdir = tempdir())
##' }
pecan.ma <- function(trait.data, prior.distns,
                     taupriors,
                     j.iter,
                     outdir,
                     random = FALSE, overdispersed = TRUE,
                     logfile = file.path(outdir, "meta-analysis.log)"),
                     verbose = TRUE) {

  mcmc.object <- list()  #  initialize output list of mcmc objects for each trait
  mcmc.mat <- list()

  ## Set inputs for jags.model()
  j.chains <- 4
  j.iter   <- as.numeric(j.iter)  # Added by SPS 08.27.2013. issue #1803
  ## log the mcmc chain parameters
  if (!is.null(logfile)) {
    sink(file = file.path(outdir, "meta-analysis.log"), split = TRUE)
    on.exit(sink(NULL), add = TRUE)
  }
  if (verbose) {
    cat(paste0("Each meta-analysis will be run with: \n", j.iter,
               " total iterations,\n", j.chains,
               " chains, \n", "a burnin of ", j.iter / 2, " samples,\n",
               ", \nthus the total number of samples will be ", j.chains * (j.iter / 2), "\n"))
  }

  for (trait.name in names(trait.data)) {

    prior <- prior.distns[trait.name, c("distn", "parama", "paramb", "n")]

    if (verbose) {
      writeLines(paste("################################################"))
      writeLines(paste("------------------------------------------------"))
      writeLines(paste("starting meta-analysis for:\n\n", trait.name, "\n"))
      writeLines(paste("------------------------------------------------"))
    }
    data <- trait.data[[trait.name]]
    data <- data[, which(!colnames(data) %in% c("cite", "trait_id", "se",
                                                "greenhouse", "site_id", "treatment_id", "trt_name", "trt_num"))]  ## remove citation and other unneeded columns


    ## check for excess missing data

    if (all(is.na(data[["obs.prec"]]))) {
      PEcAn.logger::logger.warn("NO ERROR STATS PROVIDED\n Check meta-analysis Model Convergence", 
                  "and consider turning off Random Effects by", 
                  "setting <random.effects>FALSE</random.effects>",
                  "in your pecan.xml settings file ")
    }

    if (!random) {
      data[["site"]] <- rep(1, nrow(data))
      data[["trt"]]  <- rep(0, nrow(data))
    }

    # print out some data summaries to check
    if (verbose) {
      writeLines(paste0("prior for ", trait.name, "
                     (using R parameterization):\n", prior$distn,
                     "(", prior$parama, ", ", prior$paramb, ")"))
      writeLines(paste("data max:", max(data$Y, na.rm = TRUE),
                       "\ndata min:", min(data$Y, na.rm = TRUE),
                       "\nmean:", signif(mean(data$Y, na.rm = TRUE), 3),
                       "\nn:", length(data$Y)))
      writeLines("stem plot of data points")
      writeLines(paste(stem(data$Y)))
      if (any(!is.na(data$obs.prec)) && all(!is.infinite(data$obs.prec))) {
        writeLines("stem plot of obs.prec:")
        writeLines(paste(stem(data[["obs.prec"]] ^ 2)))
      } else {
        writeLines(paste("no estimates of SD for", trait.name))
      }
    }

    jag.model.file <- file.path(outdir, paste0(trait.name, ".model.bug"))  # file to store model

    ## run the meta-analysis in JAGS

    jags.out <- single.MA(
      data,
      j.chains,
      j.iter = j.iter,
      tauA = taupriors[["tauA"]],
      tauB = taupriors[["tauB"]][[trait.name]],
      prior = prior,
      jag.model.file = jag.model.file,
      overdispersed = overdispersed
    )

    if (verbose) {
      print(summary(jags.out))
    }
    
    jags.out.trunc <- window(jags.out, start = j.iter / 2)

    mcmc.object[[trait.name]] <- jags.out.trunc
  }

  return(mcmc.object)
} # pecan.ma
