#' @param ensemble.out Data frame containing ensemble output, as generated by `load_ensemble`
#' @param x Name of variable to plot on X axis
#' @param y Name of variable to plot on Y axis
#' @param pdfs Display probability density functions outside plots. Default=TRUE.
#' @param fit.method Method for regression fit. Either "lm" for linear OLS regression (default) or "spline" for `smooth.spline` fit.
plotEnsemble <- function(ensemble.out, x, y, pdfs=TRUE, fit.method="lm", ...){
  message('plotEnsemble x: ', x)
  message('plotEnsemble y: ', y)
  error_plot <- function(err){
    plot.new()
    text(0.5, 0.5, err[1])
  }
  if(pdfs){
    zones <- matrix(c(1, 4, 3, 2), ncol=2, byrow=TRUE)
    layout(zones, widths=c(4/5, 1/5), heights=c(1/5, 4/5))
    pdfx <- try(density(ensemble.out[,x]))
    pdfy <- try(density(ensemble.out[,y]))
    pdf_succ <- c(class(pdfx), c(class(pdfy))) != "try-error"
    if(pdf_succ[1]){
      par(mar=c(0,3,1,1))
      plot(pdfx$x, pdfx$y, type="l", lwd=3, axes = FALSE, bty = "n", xlab = "", ylab = "")
    } else {
      error_plot(pdfx)
    }
    if(pdf_succ[2]){
      par(mar=c(3,0,1,1))
      plot(pdfy$y, pdfy$x, type="l", lwd=3, axes = FALSE, bty = "n", xlab = "", ylab = "")
    } else {
      error_plot(pdfy)
    }
    par(mar=c(4,4,1,1))
  }
  form <- formula(sprintf("%s ~ %s", y, x))
  plot(form, ensemble.out, col="grey50", ...)
  if(!is.na(fit.method)){
    if(fit.method == "lm"){
      fitline <- lm(form, data=ensemble.out)
      abline(fitline, lwd=3)
    } else if(fit.method == "spline"){
      fitline <- smooth.spline(ensemble.out[,x], ensemble.out[,y])
      lines(fitline, lwd=3)
    }
    else{
      stop("Unrecognized fit function")
    }
  }
}

plotAllParams <- function(ensemble.out, variable, param_names, plot_cols = 3){
  plot_rows <- ceiling(length(param_names) / plot_cols)
  par(mfrow = c(plot_rows, plot_cols), mar=c(4, 2, 3, 1), 
      mgp = c(2, 1, 0), oma = c(0, 2, 0, 0))
  for(param in param_names){
    fit_summary <- fitSummary(ensemble.out, param, variable)
    main <- with(fit_summary, sprintf("R2 = %.3g, m = %.3g, p = %.3g", r2, coefs[2,1], coefs[2,4]))
    plotEnsemble(ensemble.out, param, variable, pdfs=FALSE, ylab=NA, main=main)
  }
  mtext(variable, 2, outer = TRUE)  
}

plotAllVars <- function(ensemble.out, param, var_names, plot_cols = 3){
  plot_rows <- ceiling(length(var_names) / plot_cols)
  par(mfrow = c(plot_rows, plot_cols), mar=c(2, 3, 2, 1), 
      mgp = c(2, 1, 0), oma = c(2, 0, 1, 0))
  for(variable in var_names){
    fit_summary <- fitSummary(ensemble.out, param, variable)
    main <- with(fit_summary, sprintf("R2 = %.3g, m = %.3g, p = %.3g", r2, coefs[2,1], coefs[2,4]))
    plotEnsemble(ensemble.out, param, variable, pdfs=FALSE, xlab=NA, main=main)
  }
  mtext(param, 1, outer = TRUE)  
}

fitSummary <- function(ensemble.out, x, y) {
  message('Ensemble out names: ', paste(names(ensemble.out), collapse = ', '))
  message('fitSummary x: ', x)
  message('fitSummary y: ', y)
  form <- formula(sprintf("%s ~ %s", y, x))
  fitline <- lm(form, data=ensemble.out)
  fit_summary <- summary(fitline)
  coefs <- fit_summary$coefficients
  r2 <- fit_summary$r.squared
  adjr2 <- fit_summary$adj.r.squared
  return(list(coefs = coefs, r2 = r2, adjr2 = adjr2))
}
